#!/usr/bin/python
# -*- coding: utf-8 -*-

# Do not forget to change version number in mkdocs.yml also!
__version__ = (2, 0, 0)
__author__ = "Rafał Furmański"
__contact__ = "r.furmanski@gmail.com"
__homepage__ = "http://github.com/r4fek/django-cassandra-engine"
__docformat__ = "markdown"

default_app_config = 'django_cassandra_engine.apps.AppConfig'
